/*
 * Copyright (c) 1995, 2010, 2011 Freescale Semiconductor, Inc.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY Freescale Semiconductor ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL Freescale Semiconductor BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "ppc-asm.h"

	.file	"memcpy.S"
	.text
	.globl	memcpy
	.type	memcpy,@function

memcpy:
	li r0, 0
	dcbt	0, r4		//Initiate source prefetch even before parameter validations

	cmplwi  cr1, r5, 1
	beq		cr1, SIMPLE_1_BYTE	//only copy 1 byte, skip the overheads
	bltlr-	cr1

	addi    r8, r3, 0	// Copy destination, we do this to preserve the return which is
	
	//Touch the cache line for store as well as load in case of small copy operations
	dcbtst	0, r3

	// If source < destination we must reverse copy to handle possible overlap
	cmplw  	cr1, r4, r3
//	blt 	cr1, REVERSE // Note: ISO C Library standard define overlapping memcpy behavior as undefined.
                         // for performance reason we do not handle overlap buffer in optimize case.
	beqlr-	cr1


	// Test if source and destination are double aligned
	rlwinm 	r9, r4, 0, 29, 31	 // r9 = r4 & 7
	rlwinm 	r10, r3, 0, 29, 31	 // r10 = r3 & 7

	// Test if source and destination are the same distance from alignment
	xor.	r11, r10, r9	// (.) Enables record mode. Tests target vs 0 and sets CR0
	bne-   	cr0, TEST_WORD_ALIGNMENT // Try word next


	//  Test if they are double aligned
	andi.	r11, r10, 7	      // If (r10 & 7 == 0), then goto DOUBLE_COPY_SETUP
	beq+    DOUBLE_COPY_SETUP //(+) Assume branch will be taken

	// Here source and destination are off by the same distance.
	// Align for double word copy after
	// We can never be more then 7 bytes off double word (8 byte) alignment
	// Therefore 8 - ( source & 0x07) = distance from alignment
	subfic 	r6, r9, 8
	mtcrf   0x1, r5
	subf.	r5, r6, r5	// n -= 8 - (source & 0x7)

	// Copy bytes until both source and destination are double word aligned
	ble 	ALIGN_BYTE
	mtcrf 	0x1, r6

ALIGN_BYTE:
	sub	  	r6, r4, r8
	bf 		31,ALIGN_HALF
	lbzx 	r7, r6, r8
	stb 	r7,0(r8)
	addi 	r8, r8, 1

ALIGN_HALF:
	bf 		30,ALIGN_WORD
	lhzx 	r7, r6, r8
	sth 	r7, 0(r8)
	addi 	r8, r8, 2

ALIGN_WORD:
	bf 		29,ALIGN_DOUBLE
	lwzx 	r7, r6, r8
	stw 	r7,0(r8)
	addi 	r8, r8, 4

ALIGN_DOUBLE:
	blelr	// nothing more to do
	add 	r4, r8, r6 		// just update r4

DOUBLE_COPY_SETUP:
	srwi.   r6, r5, 5	// n / 32
    beq     SMALL_COPY_31

	//  Check to see if length to be copied is atleast 2 cache lines
	srwi.   r7, r5, 7	// n / 128
	bne 	DW_CACHE_ALIGN_SETUP

	dcbt	0, r4
	mtctr   r6              // Set count register for double words to copy

.align 4
DOUBLE_LOOP:
	dcbtst	0, r8
	lfd 	f1, 0(r4)
	lfd     f2, 8(r4)
	lfd     f3, 16(r4)
	lfd     f4, 24(r4)
	addi    r4, r4, 32
	dcbt	0, r4
	stfd    f1, 0(r8)
	stfd    f2, 8(r8)
	stfd    f3, 16(r8)
	stfd    f4, 24(r8)
	addi    r8, r8, 32

	bdnz    DOUBLE_LOOP

	andi.   r5, r5, 31	// Number of bytes remaining after unrolled loop
	beqlr   cr0		// Return now if done (r6 == 0)
	b		SMALL_COPY_31

DW_CACHE_ALIGN_SETUP:
	// We will align the destination address to 64 byte boundary 
	rlwinm. r10, r8, 0, 26, 31	 // r10 = r3 & 0x3f
	beq		DW_CACHE_COPY_SETUP

	subfic 	r6, r10, 64
	subf	r5, r6, r5	// n -= 64 - (source & 0x3f)

	srwi	r6, r6, 3 	// n / 8
	mtctr 	r6
	
// Optimize it further
DW_CACHE_ALIGN_LOOP:
	lfd 	f1, 0(r4)
	addi 	r4, r4, 8
	stfd    f1, 0(r8)
	addi 	r8, r8, 8

	bdnz   	DW_CACHE_ALIGN_LOOP 
	

DW_CACHE_COPY_SETUP:
	dcbt	0, r4
	srwi   	r6, r5, 6 	// n / 64

	addi	r7, r4, 64	//Next cache line
	dcbt	0, r7

	mtctr   r6              // Set count register for double words to copy
	
DW_CACHE_COPY_LOOP:
	dcbzl	0, r8
	addi	r7, r7, 64	//Next to next cache line
	dcbt	0, r7

	lfd 	f1, 0(r4)
	lfd     f2, 8(r4)
	lfd     f3, 16(r4)
	lfd     f4, 24(r4)

	stfd    f1, 0(r8)
	stfd    f2, 8(r8)
	stfd    f3, 16(r8)
	stfd    f4, 24(r8)

	lfd 	f1, 32(r4)
	lfd     f2, 40(r4)
	lfd     f3, 48(r4)
	lfd     f4, 56(r4)

	addi    r4, r4, 64

	stfd    f1, 32(r8)
	stfd    f2, 40(r8)
	stfd    f3, 48(r8)
	stfd    f4, 56(r8)

	addi    r8, r8, 64

	bdnz   	DW_CACHE_COPY_LOOP 
	andi.   r5, r5, 63 	// Number of bytes remaining after unrolled loop
	beqlr   cr0		// Return now if done (r6 == 0)

// Not optimized
SMALL_COPY_63:
SMALL_COPY_8x4:
	mtcrf	0x2, r5
	mtcrf	0x1, r5
	sub		r6, r4, r8
	bf 		26,MY_SMALL_COPY_16
	lfd 	f1, 0(r4)
	lfd		f2, 8(r4)
	lfd 	f3, 16(r4)
	lfd		f4, 24(r4)
	addi 	r4, r4, 32
	stfd 	f1, 0(r8)
	stfd	f2, 8(r8)
	stfd 	f3, 16(r8)
	stfd	f4, 24(r8)
	addi 	r8, r8, 32
	b 		MY_SMALL_COPY_16
	

	// Clean up remaining bytes to copy
	//optimize edge case
SMALL_COPY_31:
SMALL_COPY_8x2:
	mtcrf	0x2, r5
	mtcrf	0x1, r5
	sub		r6, r4, r8

MY_SMALL_COPY_16:
	bf 		27,SMALL_COPY_8
	lfd 	f1, 0(r4)
	lfd		f2, 8(r4)
	stfd 	f1, 0(r8)
	stfd	f2, 8(r8)
	addi 	r8, r8, 16

SMALL_COPY_8:
	bf 		28,SMALL_COPY_4
	lfdx 	f1, r6, r8
	stfd 	f1, 0(r8)
	addi 	r8, r8, 8

SMALL_COPY_4:
	bf 		29,SMALL_COPY_2
	lwzx 	r7, r6, r8
	stw 	r7,0(r8)
	addi 	r8, r8, 4

SMALL_COPY_2:
	bf 		30,SMALL_COPY_1
	lhzx 	r7, r6, r8
	sth 	r7, 0(r8)
	addi 	r8, r8, 2

SMALL_COPY_1:
	bflr 	31
	lbzx 	r7, r6, r8
	stb 	r7,0(r8)

	blr			// Return (dst)

TEST_WORD_ALIGNMENT:
	rlwinm 	r9, r4, 0, 30, 31	 // r7 = r4 & 3
	rlwinm	r10, r3, 0, 30, 31	 // r10 = r3 & 3

	xor.	r11, r10, r9         // Are they the same distance from alignment

	bne-	cr0, TEST_HW_ALIGNMENT	//Try half word now

#if 0
	bne-    cr0, BYTE_SETUP  	 // Fail here src & dst are misaligned not by the same distance
#endif

	//  Test if they are word aligned
	andi.	r11, r10, 3	      // If (r10 & 3 == 0), then goto WORD_COPY_SETUP
	beq+    WORD_COPY_SETUP //(+) Assume branch will be taken


	// Here source and destination are off by the same distence.
	// Align for word copy after
	subfic	r9, r9, 4 	//  Update the number of bytes we copied to align
	mtcrf   0x1, r5
	subf.	r5, r9, r5

	// Byte copy to align on word boundary
	ble 	ALIGN_BYTE2
	mtcrf 	0x1, r9

ALIGN_BYTE2:
	sub		r6, r4, r8
	bf		31,ALIGN_HALF2
	lbzx	r7, r6, r8
	stb		r7,0(r8)
	addi	r8, r8, 1

ALIGN_HALF2:
	bf		30,ALIGN_WORD2
	lhzx	r7, r6, r8
	sth		r7, 0(r8)
	addi	r8, r8, 2

ALIGN_WORD2:
	blelr	// nothing more to do
	add 	r4, r8, r6 		// just update r4

WORD_COPY_SETUP:
	srwi.  	r6, r5, 4	// n / 16
	beq     SMALL_COPY_15

	//  Check to see if length to be copied is atleast 2 cache lines
	srwi.   r7, r5, 7	// n / 128
	bne 	WORD_CACHE_ALIGN_SETUP

	addi    r4, r4, -4
	addi    r8, r8, -4
	mtctr   r6           

.align 4
WORD_LOOP:        
	lwz     r9, 4(r4)
	lwz     r10, 8(r4)
	lwz     r11, 12(r4)
	lwzu	r12, 16(r4)
	stw     r9, 4(r8)
	stw     r10, 8(r8)
	stw     r11, 12(r8)
	stwu	r12, 16(r8)
	bdnz    WORD_LOOP

	andi. 	r5, r5, 15	// Number of bytes remaining to copy from unrolled word loop
	beqlr   cr0             // If 0 return (dst)
	addi    r4, r4, 4
	addi    r8, r8, 4

//optimize edge case
SMALL_COPY_15:
	cmpli	0, 0, r5, 15
	bne+	SMALL_COPY_4x2
	lbzx	r6, r5, r8
	lwz 	r7, 0(r4)
	lwz     r9, 4(r4)
	lwz     r10, 8(r4)
	lwz     r11, 12(r4)
	stw		r7, 0(r8)
	stw		r9, 4(r8)
	stw		r10, 8(r8)
	stw		r11, 12(r8)
	stbx	r6, r5, r8
	blr

SMALL_COPY_4x2:
	mtcrf	0x1, r5
	sub		r6, r4, r8

	bf 		28,SMALL_COPY_4x1
	lwz 	r7, 0(r4)
	lwz 	r9, 4(r4)
	stw 	r7, 0(r8)
	stw 	r9, 4(r8)
	addi 	r8, r8, 8

SMALL_COPY_4x1:
	bf 		29,SMALL_COPY_2x1
	lwzx 	r7, r6, r8
	stw 	r7,0(r8)
	addi 	r8, r8, 4

SMALL_COPY_2x1:
	bf 		30,SMALL_COPY_1x1
	lhzx 	r7, r6, r8
	sth 	r7, 0(r8)
	addi 	r8, r8, 2

SMALL_COPY_1x1:
	bflr 	31
	lbzx 	r7, r6, r8
	stb 	r7,0(r8)

	blr			// Return (dst)

// Byte Copy - Below we copy bytes only. Here we have a source and destinations with
// 		differing alignments. 
BYTE_SETUP:
	srwi.   r7, r5, 7	// n / 128
	bne 	BYTE_CACHE_ALIGN_SETUP

	addi 	r7, r4, -1 
	addi    r8, r3, -1	// Using update loads and stores for pre-decrement
	mtctr   r5              // Move number of bytes to count register

.align 4
BYTE_LOOP_3:        
	lbzu    r9, 1(r7)
	stbu    r9, 1(r8)
	bdnz    BYTE_LOOP_3

	blr			// Return (dst)


// Byte Copy - Below we copy bytes only. Here we have a source and destinations with
// 		differing alignments. 
BYTE_CACHE_ALIGN_SETUP:
	rlwinm. r10, r8, 0, 26, 31	 // r10 = r3 & 0x3f
	beq+	BYTE_COPY_SETUP

	subfic 	r6, r10, 64
	subf.	r5, r6, r5	// n -= 64 - (source & 0x3f)

	mtctr 	r6
	addi 	r7, r4, -1		// Decrement src and destination addr by 1 as we would sue lbzu and stbu instructions
	addi 	r9, r8, -1
B_ALIGN_LOOP:
	lbzu    r10, 1(r7)
	stbu    r10, 1(r9)
	bdnz    B_ALIGN_LOOP

	add 	r4, r4, r6
	add 	r8, r8, r6

BYTE_COPY_SETUP:
	dcbt 	0, r4
	addi 	r10, r4, 64
	dcbt 	0, r10		//Next cache line

	srwi   	r7, r5, 6 	// n / 64
	

.align 4
BLOCK64_BYTE_LOOP:        
	addi 	r10, r4, 128	//Next to next cache line
	dcbt 	0, r10

	li 		r11, 4
	mtctr 	r11

	dcbzl 0, r8

	//and	r10, r10, r10

BYTE64_COPY_LOOP:
	lwz		r9, 0(r4)
	lwz		r10, 4(r4)
	lwz		r11, 8(r4)
	lwz		r12, 12(r4)
	addi	r4, r4, 16
	stw		r9, 0(r8)
	stw		r10, 4(r8)
	stw		r11, 8(r8)
	stw		r12, 12(r8)
	addi	r8, r8, 16
	bdnz 	BYTE64_COPY_LOOP

	//and	r11, r11, r11

	addi 	r7, r7, -1
	cmpi  	0,r7, 0
	bne  	BLOCK64_BYTE_LOOP	

	andi.   r5, r5, 63 	// Number of bytes remaining after unrolled loop
	beqlr   cr0		// Return now if done (r6 == 0)

	addi 	r7, r4, -1 
	addi    r8, r8, -1	// Using update loads and stores for pre-decrement
	mtctr   r5              // Move number of bytes to count register

.align 4
BYTE_LOOP_5:        
	lbzu    r9, 1(r7)
	stbu    r9, 1(r8)
	bdnz    BYTE_LOOP_5

	blr			// Return (dst)

WORD_CACHE_ALIGN_SETUP:
	// We will align the destination address to 64 byte boundary 
	rlwinm. r10, r8, 0, 26, 31	 // r10 = r3 & 0x3f
	beq+	WORD_BDRY_4DW_COPY_SETUP

	subfic 	r6, r10, 64
	subf.	r5, r6, r5	// n -= 64 - (source & 0x3f)

	// Since we are at word boundary, copy 4 bytes to reach the cache line alignment -- 64 bytes
	srwi.   r6, r6, 2 	// n / 4
	mtctr 	r6
	
// Optimize it further
WORD_BDRY_CACHE_ALIGN_LOOP:
	lwz     r9, 0(r4)
	addi 	r4, r4, 4
	stw     r9, 0(r8)
	addi 	r8, r8, 4

	bdnz   	WORD_BDRY_CACHE_ALIGN_LOOP 
	

WORD_BDRY_4DW_COPY_SETUP:
	dcbt	0, r4
	srwi.   r6, r5, 6 	// n / 64
    beq-    SMALL_COPY_WORD_63

	addi    r7, r4, 64
	dcbt    0, r7		//next cache line

	mtctr   r6              // Set count register for double words to copy
	
.align 4
WORD_BDRY_4DW_LOOP:
	// TODO Optimize the placements of dcbzl and dcbt
	// Do a dcbt for next cache line 
	addi	r7, r7, 64	
	dcbt	0, r7		//next to next cache line
	
	dcbzl    0, r8


	lwz     r9, 0(r4)
	lwz     r10, 4(r4)
	lwz     r11, 8(r4)
	lwz		r12, 12(r4)

	addi	r4, r4, 16

	stw     r9, 0(r8)
	stw     r10, 4(r8)
	stw     r11, 8(r8)
	stw		r12, 12(r8)

	addi	r8, r8, 16

	lwz     r9, 0(r4)
	lwz     r10, 4(r4)
	lwz     r11, 8(r4)
	lwz		r12, 12(r4)

	addi	r4, r4, 16

	stw     r9, 0(r8)
	stw     r10, 4(r8)
	stw     r11, 8(r8)
	stw		r12, 12(r8)

	addi	r8, r8, 16


	lwz     r9, 0(r4)
	lwz     r10, 4(r4)
	lwz     r11, 8(r4)
	lwz		r12, 12(r4)

	addi	r4, r4, 16

	stw     r9, 0(r8)
	stw     r10, 4(r8)
	stw     r11, 8(r8)
	stw		r12, 12(r8)

	addi	r8, r8, 16

	lwz     r9, 0(r4)
	lwz     r10, 4(r4)
	lwz     r11, 8(r4)
	lwz		r12, 12(r4)

	addi	r4, r4, 16

	stw     r9, 0(r8)
	stw     r10, 4(r8)
	stw     r11, 8(r8)
	stw		r12, 12(r8)
	addi	r8, r8, 16

	bdnz   	WORD_BDRY_4DW_LOOP 
	andi.   r5, r5, 63 	// Number of bytes remaining after unrolled loop
	beqlr   cr0		// Return now if done (r6 == 0)

SMALL_COPY_WORD_63:
	mtcrf	0x2, r5
	mtcrf	0x1, r5
	sub		r6, r4, r8
	bf 		26,SMALL_COPY_WORD_16
	// copy 32 bytes wordwise
	lwz     r9, 0(r4)
	lwz     r10, 4(r4)
	lwz     r11, 8(r4)
	lwz		r12, 12(r4)
	addi	r4, r4, 16 
	stw     r9, 0(r8)
	stw     r10, 4(r8)
	stw     r11, 8(r8)
	stw		r12, 12(r8)
	addi	r8, r8, 16 

	lwz     r9, 0(r4)
	lwz     r10, 4(r4)
	lwz     r11, 8(r4)
	lwz		r12, 12(r4)
	addi	r4, r4, 16 
	stw     r9, 0(r8)
	stw     r10, 4(r8)
	stw     r11, 8(r8)
	stw		r12, 12(r8)
	addi	r8, r8, 16 
	

SMALL_COPY_WORD_16:
	bf 		27,SMALL_COPY_WORD_8
	// copy 32 bytes
	lwz     r9, 0(r4)
	lwz     r10, 4(r4)
	lwz     r11, 8(r4)
	lwz		r12, 12(r4)
	addi	r4, r4, 16 

	stw     r9, 0(r8)
	stw     r10, 4(r8)
	stw     r11, 8(r8)
	stw		r12, 12(r8)
	addi	r8, r8, 16 

SMALL_COPY_WORD_8:
	bf 		28,SMALL_COPY_WORD_4
	// copy 8 bytes
	lwz 	r7, 0(r4)
	lwz 	r9, 4(r4)
	stw 	r7, 0(r8)
	stw 	r9, 4(r8)
	addi 	r8, r8, 8

SMALL_COPY_WORD_4:
	bf 		29,SMALL_COPY_WORD_2
	// copy 4 bytes
	lwzx 	r7, r6, r8
	stw 	r7,0(r8)
	addi 	r8, r8, 4

SMALL_COPY_WORD_2:
	bf 		30,SMALL_COPY_WORD_1
	lhzx 	r7, r6, r8
	sth 	r7, 0(r8)
	addi 	r8, r8, 2

SMALL_COPY_WORD_1:
	bflr 	31
	lbzx 	r7, r6, r8
	stb 	r7,0(r8)

	blr			// Return (dst)

TEST_HW_ALIGNMENT:
	rlwinm	r9, r4, 0, 31, 31
	rlwinm	r10, r3, 0, 31, 31

	xor.	r11, r10, r9		//Same distance from alignment
	bne-	cr0, BYTE_SETUP		//src & dst are not same distance from half word

	srwi.   r7, r5, 7	// n / 128
	bne 	HW_CACHE_ALIGN_SETUP

	andi.	r7, r5, 127	//Number of bytes
	srwi	r7, r5, 1	//Number of half words

	mtctr   r7              // Move number of half words to count register

HW_LOOP:
	lhz		r6, 0(r4)
	addi	r4, r4, 2
	sth		r6, 0(r8)
	addi	r8, r8, 2
	bdnz	HW_LOOP

	andi.	r5, r5, 1
	beqlr   cr0		// Return now if done

	/*todo, copy last byte*/
	lbzu	r6, 0(r4)
	stbu	r6, 0(r8)
	blr
	

HW_CACHE_ALIGN_SETUP:
	rlwinm. r10, r8, 0, 26, 31	 // r10 = r3 & 0x3f
	beq		HW_CACHE_COPY_SETUP

	subfic 	r6, r10, 64	//r6 = remaining bytes till next cache line
	mtcrf   0x1, r5
	subf.	r5, r6, r5	//r5 = remaining bytes after alignment
	mtcrf 	0x1, r6
	srwi	r6, r6, 1 	//r6 = no of half words till next cache line
	mtctr 	r6

HW_ALIGN: // Byte copy to align on halfword boundary
//	sub		r6, r4, r8
	bf		31,HW_CACHE_ALIGN_LOOP
	lbz		r7,0(r4)
	stb		r7,0(r8)
	addi	r8, r8, 1
	addi	r4, r4, 1
	
HW_CACHE_ALIGN_LOOP:
	lhz	r6, 0(r4)
	addi	r4, r4, 2
	sth	r6, 0(r8)
	addi	r8, r8, 2

	bdnz	HW_CACHE_ALIGN_LOOP

HW_CACHE_COPY_SETUP:
	li		r11, 64
	srwi   	r6, r5, 6 	// r6 = number of cache lines
	mtctr	r6
	dcbt	r11, r4		//Next cache line

	/*Loading first 8 bytes*/
	lhz		r6, 0(r4)

	lwz		r7, 2(r4)

	lhz		r9, 6(r4)
	addi	r4, r4, 8


HW_CACHE_COPY_LOOP:
	li		r11, 128
	dcbzl	0, r8

	dcbt	r11, r4		//Next to next cache line

	//6, 7, 9 (10, 12, 11)			<- 0 bytes copies

	rlwinm	r12, r6, 16, 0, 15
	lhz 	r10, 0(r4)			

	rlwinm	r11, r7, 16, 16, 31
	lwz		r6, 2(r4)			

	or		r12, r12, r11

	stw		r12, 0(r8)			
	rlwinm	r11, r7, 16, 0, 15

	lhz		r7, 6(r4)			
	or		r12, r11, r9

	stw		r12, 4(r8)			

	//10, 6, 7 (9, 11, 12)			<- 8 bytes copied
	rlwinm	r11, r10, 16, 0, 15

	lhz		r9, 8(r4)
	rlwinm	r12, r6, 16, 16, 31

	lwz		r10, 10(r4)
	or		r11, r11, r12

	stw		r11, 8(r8)
	rlwinm	r12, r6, 16, 0, 15

	lhz		r6, 14(r4)
	or		r11, r12, r7

	stw		r11, 12(r8)

	//9, 10, 6 (7, 12, 11)			<- 16 bytes copied
	rlwinm	r12, r9, 16, 0, 15

	lhz		r7, 16(r4)
	rlwinm	r11, r10, 16, 16, 31

	lwz		r9, 18(r4)
	or		r12, r12, r11

	stw		r12, 16(r8)
	rlwinm	r11, r10, 16, 0, 15

	lhz		r10, 22(r4)

	or		r12, r11, r6
	stw		r12, 20(r8)


	//7, 9, 10 (6, 11, 12)			<- 24 bytes copied

	rlwinm	r11, r7, 16, 0, 15
	lhz		r6, 24(r4)

	rlwinm	r12, r9, 16, 16, 31
	lwz		r7, 26(r4)

	or		r11, r12, r11
	stw		r11, 24(r8)

	rlwinm	r12, r9, 16, 0, 15
	lhz		r9, 30(r4)

	or		r11, r12, r10
	stw		r11, 28(r8)


	//6, 7, 9 (10, 11, 12)			<- 32 bytes copied

	rlwinm	r12, r6, 16, 0, 15
	lhz 	r10, 32(r4)			

	rlwinm	r11, r7, 16, 16, 31
	lwz		r6, 34(r4)			

	or		r12, r12, r11
	stw		r12, 32(r8)			

	rlwinm	r11, r7, 16, 0, 15
	lhz		r7, 38(r4)			

	or		r12, r11, r9
	stw		r12, 36(r8)		


	//10, 6, 7 (9, 11, 12)			<- 40 bytes copied

	rlwinm	r11, r10, 16, 0, 15
	lhz		r9, 40(r4)

	rlwinm	r12, r6, 16, 16, 31
	lwz		r10, 42(r4)

	or		r12, r12, r11
	stw		r12, 40(r8)

	rlwinm	r11, r6, 16, 0, 15
	lhz		r6, 46(r4)

	or		r12, r11, r7
	stw		r12, 44(r8)


	//9, 10, 6 (7, 11, 12)			<- 48 bytes copied

	rlwinm	r11, r9, 16, 0, 15
	lhz		r7, 48(r4)

	rlwinm	r12, r10, 16, 16, 31
	lwz		r9, 50(r4)

	or		r11, r12, r11
	stw		r11, 48(r8)

	rlwinm	r12, r10, 16, 0, 15
	lhz		r10, 54(r4)

	or		r11, r12, r6
	stw		r11, 52(r8)


	//7, 9, 10 (6, 11, 12)			<- 56 bytes copied

	rlwinm	r12, r7, 16, 0, 15
	lhz		r6,  56(r4)

	rlwinm	r11, r9, 16, 16, 31
	lwz		r7, 58(r4)

	or		r12, r12, r11
	stw		r12, 56(r8)

	rlwinm	r11, r9, 16, 0, 15
	lhz		r9, 62(r4)
	addi	r4, r4,64 			

	or		r12, r11, r10
	stw		r12, 60(r8)
	addi	r8, r8, 64			

	bdnz 	HW_CACHE_COPY_LOOP
copy_complete:

	andi.   r5, r5, 63 	// Number of bytes remaining after unrolled loop
	beqlr   cr0		// Return now if done (r5 == 0)

	addi	r8, r8, -2
	addi	r4, r4, -10	//8 bytes have been loaded extra

	srwi.	r6, r5, 1	//r6 = r5/2
	beq		HW_LASTBYTE
	mtctr   r6              // Move number of half words to count register

HW_LOOP1:
	lhzu	r6, 2(r4)
	sthu	r6, 2(r8)
	bdnz	HW_LOOP1

	andi.   r5, r5, 1 	// Number of bytes remaining after unrolled loop
	beqlr   cr0		// Return now if done (r5 == 0)

HW_LASTBYTE:
	lbzu	r6, 2(r4)
	stbu	r6, 2(r8)
	blr

SIMPLE_1_BYTE:
	lbz     r9, 0(r4)
	stb     r9, 0(r3)
	blr
