/*
 * Copyright (c) 2011 Freescale Semiconductor, Inc.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY Freescale Semiconductor ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL Freescale Semiconductor BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <stdint.h>
#include <stddef.h>
#include <string.h>

#define CACHE_LINE_SIZE		64

#define dcbzl(p) \
	do { \
		__asm__ __volatile__ ("dcbzl 0,%0" : : "r" (p)); \
	} while(0)

#if __WORDSIZE == 64
#define set_cl_long(p_cl, w) { \
	p_cl[0] = w;	\
	p_cl[1] = w;	\
	p_cl[2] = w;	\
	p_cl[3] = w;	\
	p_cl[4] = w;	\
	p_cl[5] = w;	\
	p_cl[6] = w;	\
	p_cl[7] = w;	\
}
#elif __WORDSIZE == 32
#define set_cl_long(p_cl, w) { \
	p_cl[0] = w;	\
	p_cl[1] = w;	\
	p_cl[2] = w;	\
	p_cl[3] = w;	\
	p_cl[4] = w;	\
	p_cl[5] = w;	\
	p_cl[6] = w;	\
	p_cl[7] = w;	\
	p_cl[8] = w;	\
	p_cl[9] = w;	\
	p_cl[10] = w;	\
	p_cl[11] = w;	\
	p_cl[12] = w;	\
	p_cl[13] = w;	\
	p_cl[14] = w;	\
	p_cl[15] = w;	\
}
#else
#error "Unknown __WORDSIZE" 
#endif

void * memset(void * dst, int val, size_t n)
{
	unsigned long *p_ul;
	unsigned long count;
	unsigned long v;
	size_t rem;
	uint8_t *p_uc;
	uint8_t i;
    
	p_uc = dst;

	__builtin_prefetch(p_uc, 1);

	/* Unlikely case when n = 0 is automatically handled. */
	i = (uint8_t)val;	/*Truncate extra bytes if any*/
	
	if (n < (2 * sizeof(*p_ul))) {
		rem = n;
		goto end_byte_store;
	}

	/*Find bytes to copy till long boundary*/
	rem = (~((unsigned long)p_uc - 1) & (sizeof(*p_ul) - 1));
	n = n - rem;

	/*Now reach long boundary*/
	while (rem) {
		*p_uc = i;
		p_uc++;
		rem--;
	}
	
	v = (unsigned long)(i << 24 | i << 16 | i <<  8 | i);
#if __WORDSIZE == 64
	v = (v << 32 | v);
#endif
	p_ul = (unsigned long *)p_uc;
	
	if (n < (2 * CACHE_LINE_SIZE)) {
		rem = n;
		goto end_long_store;
	}
	
	/*How far are we from cache line boundary*/
	rem = (~((unsigned long)p_uc - 1) & (CACHE_LINE_SIZE - 1));
	
	/*Set long words till cache boundary*/
	count = rem / sizeof(*p_ul);
	n = n - rem;
	while (count) {
		*p_ul = v;
		p_ul++;
		count--;
	}
	
	count = n / CACHE_LINE_SIZE;
	rem = n % CACHE_LINE_SIZE;

	/*Now here we have cache line aligned address*/
	if (count) {
		if (val) {
			dcbzl(p_ul);
			while (count > 1) {
				dcbzl(p_ul + CACHE_LINE_SIZE/sizeof(*p_ul));
				set_cl_long(p_ul, v);
				p_ul = p_ul + CACHE_LINE_SIZE/sizeof(*p_ul);
				count--;
			};
			
			set_cl_long(p_ul, v);
			p_ul = p_ul + CACHE_LINE_SIZE/sizeof(*p_ul);
		} else {
			do {
				dcbzl(p_ul);
				p_ul = p_ul + CACHE_LINE_SIZE/sizeof(*p_ul);
				count--;
			} while (count);
		}
	}
	
	/* Here, the destination address is already long aligned.
	 * Hence we can use long stores directly.
	 */
end_long_store:	
	count = rem / sizeof(*p_ul);
	while (count) {
		*p_ul = v;
		p_ul++;
		count--;
	}
	
	rem = rem % sizeof(*p_ul);
	p_uc = (uint8_t *)p_ul;

end_byte_store:
	while (rem) {
		*p_uc = i;
		p_uc++;
		rem--;
	}
	
	return dst;
}

void bzero(void *s, size_t n)
{
	memset(s, 0, n);
}